import numpy as np
import time
import itertools
import random
from sys import maxsize
import sc2
from sc2 import game_info, game_data, game_state, Race
from sc2.unit import Unit
from sc2.units import Units
from sc2.ids.unit_typeid import UnitTypeId
from sc2.position import Point2, Point3
from sc2.ids.ability_id import AbilityId as AbilID
from sc2.ids.unit_typeid import UnitTypeId as UnitID
from sc2.ids.buff_id import BuffId


class Terra(sc2.BotAI):
    def select_target(self):
        target = self.enemy_structures
        if target.exists:
            return target.closest_to(self.start_location).position

        target = self.enemy_units
        if target.exists:
            return target.closest_to(self.start_location).position

        if len(self.enemy_structures) + len(self.enemy_units) < 1:
            return self.army_target.position

    def __init__(self, bot = None):

        self.bot = bot
        self.player_id = Terra
        self.attack = 0
        self.ITERATIONS_PER_MINUTE = 165
        self.MAX_WORKERS = 80
        self.currentVersion = "1.1" # update every big change or update to ladder
        self.ordered_expansions = None
        self.enemy_ordered_expansions = None
        self.clear_map = None
        self.scouts_and_spots = {}
        self.offense_amount = 10

    async def on_start(self):

        self._client.game_step = 2  # set higher for bot to have less APM/take less resources
        
        quotes = ['He who knows when to fight and when not to fight, holds the key.', 'In preparing for battle I have always found that plans are useless, but planning is indispensable.',
                  'Attack him where he is unprepared, appear where you are not expected', 'If you know neither the enemy nor yourself, you will succumb in every battle',
                  'The difficulty of tactical maneuvering consists in turning the devious into the direct, and misfortune into gain.',
                  'He who wishes to fight must first count the cost', '']

        self.ordered_expansions = sorted(self.expansion_locations.keys(), key=lambda e: e.distance_to(self.start_location))
        self.enemy_ordered_expansions = sorted(self.expansion_locations.keys(), key=lambda e: e.distance_to(self.enemy_start_locations[0]))
        
        message2 = str(random.choice(quotes))
        await self.chat_send(message2)
        print("Game has begun")

    async def on_step(self, iteration):

        if not self.townhalls:
            return
        elif self.structures(UnitID.NEXUS).amount < (self.time/180) or self.minerals > 800:
            if self.can_afford(UnitID.NEXUS):
                if self.already_pending(UnitID.NEXUS) <= self.structures(UnitID.NEXUS).not_ready.amount:
                    await self.expand_now(max_distance=0)
            else:
                return
        
        # Draw pathing grid
        # self.draw_pathing_grid()

        await self.distribute_workers()
        await self.fill_extractors()
        await self.abilityuse()
        await self.buildings()
        await self.buildworkers()
        await self.army()
        self.set_army_target()
        await self.armymicro()
        await self.armycontrol()
        
    def draw_pathing_grid(self):
        map_area = self._game_info.playable_area
        for (b, a), value in np.ndenumerate(self._game_info.pathing_grid.data_numpy):
            if value == 0:
                continue
                # Skip values outside of playable map area
            if not (map_area.x <= a < map_area.x + map_area.width):
                continue
            if not (map_area.y <= b < map_area.y + map_area.height):
                continue
            p = Point2((a, b))
            h2 = self.get_terrain_z_height(p)
            pos = Point3((p.x, p.y, h2))
            p0 = Point3((pos.x - 0.25, pos.y - 0.25, pos.z + 0.25)) + Point2((0.5, 0.5))
            p1 = Point3((pos.x + 0.25, pos.y + 0.25, pos.z - 0.25)) + Point2((0.5, 0.5))
            # print(f"Drawing {p0} to {p1}")
            color = Point3((255, 0, 0))
            self._client.debug_box_out(p0, p1, color=color)

    async def abilityuse(self):
        if self.enemy_units:
            enemy_workers = self.enemy_units.filter(
                lambda unit: unit.type_id in ({UnitID.PROBE, UnitID.SCV, UnitID.DRONE})
            )

            for oracle in self.units(UnitID.ORACLE).filter(lambda unit: unit.type_id in {UnitID.ORACLE}):
                in_range_enemys = enemy_workers.in_attack_range_of(oracle)
                if in_range_enemys and self.units(UnitID.ORACLE).ready.first.energy >= 25:
                    self.do(oracle(AbilID.BEHAVIOR_PULSARBEAMON))

                if self.units(UnitID.ORACLE).ready.first.energy < 25 or not in_range_enemys:
                    self.do(oracle(AbilID.BEHAVIOR_PULSARBEAMOFF))

    async def buildworkers(self):
        if self.townhalls.ready:
            nexus = self.townhalls.ready.random
            if self.supply_workers < min(self.townhalls.amount * 22, self.MAX_WORKERS) and nexus.is_idle:
                if self.can_afford(UnitID.PROBE):
                    self.do(nexus.train(UnitID.PROBE), subtract_cost=True, subtract_supply=True)

    async def expand_logic(self):
        if self.can_afford(UnitID.NEXUS) and not self.already_pending(UnitID.NEXUS):
            await self.expand_now()

    async def fill_extractors(self):
        for extractor in self.gas_buildings:
            # returns negative value if not enough workers
            if extractor.surplus_harvesters < 0:
                probes_with_no_resource = self.workers.filter(
                    lambda unit: not unit.is_carrying_resource and unit.is_collecting
                )
                if probes_with_no_resource:
                    # surplus_harvesters is negative when workers are missing
                    for n in range(-extractor.surplus_harvesters):
                        # prevent crash by only taking the minimum
                        probe = probes_with_no_resource[min(n, probes_with_no_resource.amount) - 1]
                        self.do(probe.gather(extractor))

    async def buildings(self):

        if self.vespene < 250 and self.structures(UnitID.GATEWAY).exists:
            for nexus in self.townhalls.ready:
                vgs = self.vespene_geyser.closer_than(15, nexus)
                for vg in vgs:
                    if not self.can_afford(UnitID.ASSIMILATOR):
                        break

                    worker = self.select_build_worker(vg.position)
                    if worker is None:
                            break

                    if not self.gas_buildings or not self.gas_buildings.closer_than(1, vg):
                        self.do(worker.build(UnitID.ASSIMILATOR, vg), subtract_cost=True)
                        self.do(worker.stop(queue=True))

        if self.townhalls(UnitID.NEXUS).ready:
            position = self.townhalls(UnitID.NEXUS).ready.random.position.towards(self.game_info.map_center, 7)
        
            if not self.structures(UnitID.PYLON).exists or self.supply_left < 3 and not self.already_pending(UnitID.PYLON):
                await self.build(UnitID.PYLON, near=position)

            if self.structures(UnitID.GATEWAY).ready.idle.amount < 1 and self.structures(UnitID.PYLON).ready \
                    and not self.already_pending(UnitID.GATEWAY) and self.can_afford(UnitID.GATEWAY):
                await self.build(UnitID.GATEWAY, near=position)

            if not self.structures(UnitID.CYBERNETICSCORE).ready and not self.already_pending(UnitID.CYBERNETICSCORE):
                if self.structures(UnitID.CYBERNETICSCORE).amount < 1 and self.can_afford(UnitID.CYBERNETICSCORE):
                    await self.build(UnitID.CYBERNETICSCORE, near=position)

            if not self.structures(UnitID.STARGATE).exists and self.structures(UnitID.CYBERNETICSCORE).ready \
                and self.structures(UnitID.PYLON).ready and self.can_afford(UnitID.STARGATE) and not \
                    self.already_pending(UnitID.STARGATE) and self.structures(UnitID.STARGATE).amount < 1:
                await self.build(UnitID.STARGATE, near=position)

            if self.units.of_type({UnitID.ZEALOT,UnitID.STALKER,UnitID.IMMORTAL}).ready.amount > 3 and not self.structures(UnitID.ROBOTICSFACILITY).ready.idle.exists and self.structures(UnitID.CYBERNETICSCORE).ready \
                    and self.structures(UnitID.PYLON).ready and self.can_afford(UnitID.ROBOTICSFACILITY) and not \
                    self.already_pending(UnitID.ROBOTICSFACILITY) and self.structures(UnitID.ROBOTICSFACILITY).ready.idle.amount < 1:
                await self.build(UnitID.ROBOTICSFACILITY, near=position)

            if not self.structures(UnitID.FLEETBEACON).exists and self.structures(UnitID.STARGATE).ready:
                if self.can_afford(UnitID.FLEETBEACON) and not self.already_pending(UnitID.FLEETBEACON):
                    if self.structures(UnitID.FLEETBEACON).amount == 0:
                        await self.build(UnitID.FLEETBEACON, near=position)

    async def army(self):
        if self.structures(UnitID.ROBOTICSFACILITY).ready and self.structures(UnitID.CYBERNETICSCORE).ready and self.can_afford(UnitID.IMMORTAL):
            # if self.units(UnitID.IMMORTAL).amount < self.time / 2:
            self.train(UnitID.IMMORTAL)

        if self.structures(UnitID.GATEWAY).ready.idle and self.structures(UnitID.CYBERNETICSCORE).ready and self.can_afford(UnitID.STALKER):
            self.train(UnitID.STALKER)

        # if self.structures(UnitID.GATEWAY).ready and self.structures(UnitID.CYBERNETICSCORE).ready and self.can_afford(UnitID.ADEPT):
        #    if self.units(UnitID.ADEPT).amount < 2 and self.time < 500:
        #        self.train(UnitID.ADEPT)

        if self.enemy_race.name == "Zerg":
            if self.structures(UnitID.GATEWAY).ready and self.can_afford(UnitID.ZEALOT):
                if self.units(UnitID.ZEALOT).amount < 5:
                    self.train(UnitID.ZEALOT)

        if self.structures(UnitID.STARGATE).ready.idle and self.can_afford(UnitID.ORACLE):
            if self.units(UnitID.ORACLE).amount < 3 and self.time < 300:
                self.train(UnitID.ORACLE)

        if self.structures(UnitID.STARGATE).ready.idle and self.structures(UnitID.FLEETBEACON).ready:
            if self.can_afford(UnitID.CARRIER):
                self.train(UnitID.CARRIER)

    async def armymicro(self):
        if self.units(UnitID.ORACLE).ready:
            enemy_workers = self.enemy_units.filter(
                lambda unit: unit.type_id in ({UnitID.PROBE, UnitID.SCV, UnitID.DRONE})
            )
            oracles = self.units.filter(lambda unit: unit.type_id in {UnitID.ORACLE})
            dangerous_units = self.enemy_units.filter(lambda  u: u.can_attack and u.type_id not in {UnitID.EGG, UnitID.LARVA, UnitID.OVERLORD, UnitID.OVERSEER, UnitID.PROBE, UnitID.DRONE, UnitID.SCV} )
            for oracle in oracles:
                closest_enemy = dangerous_units.closest_to(oracle)
                if dangerous_units.in_attack_range_of(oracle):
                    closest_enemy = dangerous_units.closest_to(oracle)
                    distance = oracle.ground_range + 40
                    self.do(oracle.move(closest_enemy.position.towards(oracle, distance)))

                if enemy_workers.amount >= 1 and not closest_enemy:
                    # attack if weapon not on cooldown
                    if oracle.weapon_cooldown == 0:
                        in_range_enemies = enemy_workers
                        # attack enemy with lowest hp of the ones in range
                        lowest_hp = min(in_range_enemies, key=lambda e: e.health + e.shield)
                        self.do(oracle.attack(lowest_hp))

    async def armycontrol(self):
        army = self.units(UnitID.STALKER) | self.units(UnitID.IMMORTAL) | self.units(UnitID.CARRIER)
        stalkers = self.units.filter(lambda unit: unit.type_id in {UnitID.STALKER})
        harass_adepts = self.units.filter(lambda unit: unit.type_id in {UnitID.ADEPT})
        zealots = self.units.filter(lambda unit: unit.type_id in {UnitID.ZEALOT})
        carrier = self.units.filter(lambda unit: unit.type_id in {UnitID.CARRIER})
        harass_oracles = self.units(UnitID.ORACLE)
        nexus = self.townhalls.first
        immortals = self.units.filter(lambda unit: unit.type_id in {UnitID.IMMORTAL})
        dangerous_units = self.enemy_units.filter(
            lambda u: u.can_attack and not self.enemy_structures and u.type_id not in {UnitID.EGG, UnitID.LARVA, UnitID.OVERLORD, UnitID.OVERSEER, UnitID.PROBE, UnitID.DRONE,
                                                         UnitID.SCV})
        attack_position = self.select_target()
        retreat_position = attack_position.position.towards(self.start_location, 30)

        if harass_oracles:
            for oracles in harass_oracles:
                if harass_oracles.first.energy <= 25:
                    self.do(oracles.move(self.townhalls.closest_to(oracles)))
                elif harass_oracles.first.energy >= 50:
                    self.do(oracles.attack(attack_position))

        if stalkers:
            for stalker in stalkers:
                if army.amount < self.offense_amount and not self.enemy_units:
                    if stalkers.ready:
                        rally = self.townhalls.closest_to(self.enemy_start_locations[0])
                        self.do(stalker.move(rally))

                if self.enemy_units:
                    if dangerous_units.amount < army.amount and not self.enemy_units.closer_than(30, self.townhalls.closest_to(self.enemy_units.first)):
                        self.do(stalker.attack(attack_position))

                    elif dangerous_units.amount * 2 > army.amount and not self.enemy_units.closer_than(30, self.townhalls.closest_to(self.enemy_units.first)):
                        self.do(stalker.move(retreat_position))

                    elif self.enemy_units.closer_than(30, self.townhalls.ready.first):
                        self.do(stalker.attack(self.enemy_units.first))

                    else:
                        self.do(stalker.attack(attack_position))

        if immortals:
            for immortal in immortals:
                if army.amount < self.offense_amount and not self.enemy_units.closer_than(30, self.townhalls.ready.first):
                    if immortals.ready and not self.enemy_units.closer_than(30, self.townhalls.ready.first) and stalkers.ready:
                        rally = self.townhalls.ready.closest_to(self.enemy_start_locations[0])
                        self.do(immortal.move(rally))
                if self.enemy_units:
                    if self.enemy_units.closer_than(30, self.townhalls.ready.first):
                        self.do(immortal.attack(self.enemy_units.first))

                    elif dangerous_units.amount < army.amount and not self.enemy_units.closer_than(30, self.townhalls.ready.first)and (army.amount > self.time/30 or self.supply_used > 190):
                        self.do(immortal.attack(attack_position))

        if zealots:
            for zealot in zealots:
                if army.amount < self.offense_amount and not self.enemy_units.closer_than(30, self.townhalls.ready.first) and stalkers.ready:
                    rally = self.townhalls.ready.closest_to(self.enemy_start_locations[0])
                    self.do(zealot.move(rally))

                if self.enemy_units:
                    if self.enemy_units.closer_than(30, self.townhalls.ready.first):
                        self.do(zealot.attack(self.enemy_units.first))

                    elif (army.amount > self.time/30 or self.supply_used > 190) and not self.enemy_units.closer_than(30, self.townhalls.ready.first):
                        self.do(zealot.attack(attack_position))

        if stalkers:
            for stalker in carrier:
                if army.amount < self.offense_amount and not self.enemy_units:
                    if stalkers.ready:
                        rally = self.townhalls.closest_to(self.enemy_start_locations[0])
                        self.do(stalker.move(rally))

                if self.enemy_units:
                    if dangerous_units.amount < army.amount and not self.enemy_units.closer_than(30, self.townhalls.closest_to(self.enemy_units.first)):
                        self.do(stalker.attack(attack_position))

                    elif dangerous_units.amount * 2 > army.amount and not self.enemy_units.closer_than(30, self.townhalls.closest_to(self.enemy_units.first)):
                        self.do(stalker.move(retreat_position))

                    elif self.enemy_units.closer_than(30, self.townhalls.ready.first):
                        self.do(stalker.attack(self.enemy_units.first))

                    else:
                        self.do(stalker.attack(attack_position))

    def set_army_target(self):
        # sets the next waypoint for the army in case there is nothing on the map
        # if we didnt start to clear the map already
        if not self.clear_map:
            # start with enemy starting location, then cycle through all expansions
            self.clear_map = itertools.cycle(self.enemy_ordered_expansions)
            self.army_target = next(self.clear_map)
            while self.army_target in self.owned_expansions:
                self.army_target = next(self.clear_map)
        # we can see the expansion but there seems to be nothing there, get next
        if self.units.closer_than(6, self.army_target):
            self.army_target = next(self.clear_map)
